import numpy as np
import pandas as pd

from .epochs import Epochs


def generate(n_epochs=10, n_samples=100, n_categories=2, n_channels=32):
    """Return Epochs object with fake EEG data.

    Parameters
    ----------

    n_epochs : int
        number of epochs per category to be generated
    n_samples : int
        number of samples in a single epochs
    n_categories : int
        number of levels of the categorical variable
    n_channels : int
        number of time series representing EEG channels

    Returns
    -------

    epochs : Epochs object
        Epochs object containing simulated EEG data.

    Notes
    -----

    ``n_epochs`` and ``n_categories`` interact in the sense that ``n_epochs``
    epochs are generated for each level of the categorical variable. In other
    words, the true number of epochs in the generated data is equal to
    ``n_epochs`` * ``n_categories``.

    For example, the default ``n_epochs = 10`` and ``n_categories
    = 2`` produces 20 epochs, 10 per category.
    """

    df, channels = _generate(n_epochs, n_samples, n_categories, n_channels)
    return Epochs(df, channels)


def _generate(n_epochs, n_samples, n_categories, n_channels):
    """Return Pandas DataFrame with fake EEG data, and a list of channels."""

    from . import EPOCH_ID, TIME

    total = n_epochs * n_samples * n_categories

    categories = np.array([f'cat{i}' for i in range(n_categories)])

    indices = {
        EPOCH_ID: np.repeat(np.arange(n_epochs * n_categories), n_samples),
        TIME: np.tile(np.arange(n_samples), n_epochs * n_categories),
    }

    predictors = {
        'categorical': np.tile(np.repeat(categories, n_samples), n_epochs),
        'continuous': np.random.uniform(size=total),
    }

    channels = [f'channel{i}' for i in range(n_channels)]
    eeg = {
        channel: np.random.normal(loc=0, scale=30, size=total)
        for channel in channels
    }

    data = {**indices, **predictors, **eeg}

    df = pd.DataFrame(data).set_index([EPOCH_ID, TIME]).sort_index()

    return df, channels



import numpy as np
from collections import defaultdict


def get_index_duplicates_table(df, level):
    """Return a string table of duplicate index values and their locations."""

    assert level in df.index.names

    level_values = df.index.get_level_values(level)
    dupe_mask = level_values.duplicated(keep=False)

    dupes = level_values[dupe_mask]
    dupe_indices = np.flatnonzero(dupe_mask)

    dupe_dict = defaultdict(list)
    for index, value in zip(dupe_indices, dupes):
        dupe_dict[value].append(index)

    # hardcoded padding
    msg = f'Duplicates in index level {level}:\n\n'
    msg += '{0:<20} {1}\n'.format(level, 'Locations')
    for value, locations in sorted(dupe_dict.items()):
        locations = (str(item) for item in locations)
        msg += '{0:<20} {1}\n'.format(value, ', '.join(locations))

    return msg


def get_first_group(groupby):

    first_group_name = list(groupby.groups)[0]
    first_group = groupby.get_group(first_group_name)
    return first_group



import numpy as np
import pandas as pd
from statsmodels.formula.api import ols, mixedlm
from tqdm import tqdm

from .errors import FitGridError
from .fitgrid import FitGrid
from . import tools


class Epochs:
    """Container class used for storing epochs tables and exposing statsmodels.

    Parameters
    ----------
    epochs_table : pandas DataFrame
        long form dataframe containing epochs with equal indices
    channels : list of str
        list of channel names to serve as dependent variables

    Returns
    -------
    epochs : Epochs
        epochs object

    """

    def __init__(self, epochs_table, channels='default'):

        from . import EPOCH_ID, TIME, CHANNELS

        if channels == 'default':
            channels = CHANNELS

        # channels must be a list of strings
        if not isinstance(channels, list) or not all(
            isinstance(item, str) for item in channels
        ):
            raise FitGridError('channels should be a list of strings.')

        # all channels must be present as epochs table columns
        missing_channels = set(channels) - set(epochs_table.columns)
        if missing_channels:
            raise FitGridError(
                'channels should all be present in the epochs table, '
                f'the following are missing: {missing_channels}'
            )

        if not isinstance(epochs_table, pd.DataFrame):
            raise FitGridError('epochs_table must be a Pandas DataFrame.')

        # these index columns are required for consistency checks
        for item in (EPOCH_ID, TIME):
            if item not in epochs_table.index.names:
                raise FitGridError(
                    f'{item} must be a column in the epochs table index.'
                )

        # make our own copy so we are immune to modification to original table
        table = epochs_table.copy().reset_index().set_index(EPOCH_ID)
        assert table.index.names == [EPOCH_ID]

        snapshots = table.groupby(TIME)

        # check that snapshots across epochs have equal index by transitivity
        prev_group = None
        for idx, cur_group in snapshots:
            if prev_group is not None:
                if not prev_group.index.equals(cur_group.index):
                    raise FitGridError(
                        f'Snapshot {idx} differs from '
                        f'previous snapshot in {EPOCH_ID} index:\n'
                        f'Current snapshot\'s indices:\n'
                        f'{cur_group.index}\n'
                        f'Previous snapshot\'s indices:\n'
                        f'{prev_group.index}'
                    )
            prev_group = cur_group

        if not prev_group.index.is_unique:
            raise FitGridError(
                f'Duplicate values in {EPOCH_ID} index not allowed:',
                tools.get_index_duplicates_table(table, EPOCH_ID),
            )

        # checks passed, set instance variables
        self.channels = channels
        self.table = table
        self._snapshots = snapshots
        self._epoch_index = tools.get_first_group(snapshots).index.copy()

    def _validate_LHS(self, LHS):

        # must be a list of strings
        if not (
            isinstance(LHS, list)
            and all(isinstance(item, str) for item in LHS)
        ):
            raise FitGridError('LHS must be a list of strings.')

        # all LHS items must be present in the epochs_table
        missing = set(LHS) - set(self.table.columns)
        if missing:
            raise FitGridError(
                'Items in LHS should all be present in the epochs table, '
                f'the following are missing: {missing}'
            )

    def _validate_RHS(self, RHS):

        # validate RHS
        if RHS is None:
            raise FitGridError('Specify the RHS argument.')
        if not isinstance(RHS, str):
            raise FitGridError('RHS has to be a string.')

    def distances(self):
        """Return scaled Euclidean distances of epochs from the "mean" epoch.

        Returns
        -------
        distances : pandas Series or DataFrame
            Series or DataFrame with epoch distances

        Notes
        -----
        Distances are scaled by dividing by the max.
        """

        from . import EPOCH_ID, TIME

        table = self.table.reset_index().set_index([EPOCH_ID, TIME])[
            self.channels
        ]

        n_channels = len(table.columns)
        n_epochs = len(table.index.unique(level=EPOCH_ID))
        n_samples = len(table.index.unique(level=TIME))

        assert table.values.size == n_channels * n_epochs * n_samples
        values = table.values.reshape(n_epochs, n_samples, n_channels)

        mean = values.mean(axis=0)
        diff = values - mean

        def l2_norm(data, axis=1):
            return np.sqrt((data * data).sum(axis=axis))

        # first n_samples is axis 1, then n_channels, leaving epochs
        distances_arr = l2_norm(l2_norm(diff))
        distances_arr_scaled = distances_arr / distances_arr.max()
        distances = pd.Series(distances_arr_scaled, index=self._epoch_index)

        return distances

    def run_model(self, function, channels):
        """Run an arbitrary model on the epochs.

        Parameters
        ----------
        function : Python function
            function that runs a model, see Notes below for details
        channels : list of str
            list of channels to serve as dependent variables

        Returns
        -------
        grid : FitGrid
            a FitGrid object containing the results

        Notes
        -----
        The function should take two parameters, ``data`` and ``channel``, run
        some model on the data, and return an object containing the results.
        ``data`` will be a snapshot across epochs at a single timepoint,
        containing all channels of interest. ``channel`` is the name of the
        target variable that the function runs the model against (uses it as
        the dependent variable).

        Examples
        --------
        Here's an example of a function that can be passed to ``run_model``::

            def regression(data, channel):
                formula = channel + ' ~ continuous + categorical'
                return ols(formula, data).fit()

        """

        results = {
            channel: self._snapshots.apply(function, channel=channel)
            for channel in tqdm(channels, desc='Channels: ')
        }

        return FitGrid(pd.DataFrame(results), self._epoch_index)

    def lm(self, LHS=None, RHS=None):
        """Run ordinary least squares linear regression on the epochs.

        Parameters
        ----------
        LHS : list of str, optional, defaults to all channels
            list of channels for the left hand side of the regression formula
        RHS : str
            right hand side of the regression formula

        Returns
        -------

        grid : FitGrid
            FitGrid object containing results of the regression

        """

        if LHS is None:
            LHS = self.channels

        self._validate_LHS(LHS)
        self._validate_RHS(RHS)

        def regression(data, channel):
            formula = channel + ' ~ ' + RHS
            return ols(formula, data).fit()

        return self.run_model(regression, LHS)

    def mlm(
        self, LHS=None, RHS=None, re_formula=None, vc_formula=None, groups=None
    ):
        """Run a linear mixed effects model on the epochs.

        Parameters
        ----------
        LHS : list of str, optional, defaults to all channels
            list of channels for the left hand side of the formula
        RHS : str
            fixed effects part of the mixed effect model formula
        re_formula : str
            A one-sided formula defining the variance structure of the model.
            The default gives a random intercept for each group.
        vc_formula : dict
            Formulas describing variance components.  ``vc_formula[vc]`` is the
            formula for the component with variance parameter named vc. The
            formula is processed into a matrix, and the columns of this matrix
            are linearly combined with independent random coefficients having
            mean zero and a common variance.
        groups : str
            name of the grouper column

        Returns
        -------
        grid : FitGrid
            FitGrid object containing results of the mixed effect model fitting

        Notes
        -----
        Note that ``statsmodels`` (and thus ``fitgrid``) does not support
        arbitrary crossed models. For more information see `Statsmodels
        documentation on mixed models
        <http://www.statsmodels.org/stable/mixed_linear.html>`_.

        """

        if LHS is None:
            LHS = self.channels

        self._validate_LHS(LHS)
        self._validate_RHS(RHS)

        def mixed_linear_model(data, channel):
            formula = channel + ' ~ ' + RHS
            return mixedlm(
                formula,
                data,
                re_formula=re_formula,
                vc_formula=vc_formula,
                groups=groups,
            )

        return self.run_model(mixed_linear_model, LHS)

    def plot_averages(self, channels=None, negative_up=True):
        """Plot grand mean averages for each channel, negative up by default.

        Parameters
        ----------
        channels : list of str, optional, default CHANNELS
            list of channel names to plot the averages
        negative_up : bool, optional, default True
            by convention, ERPs are plotted negative voltage up
        """

        if channels is None:
            channels = self.channels

        from . import plots

        data = self._snapshots.mean()
        plots.stripchart(data[channels], negative_up=negative_up)



# to install in development mode, from this directory run
#
#     pip install --user -e .
#
# to install stable package systemwide, as root run
#
#     pip install .
#
# http://python-packaging.readthedocs.io/en/latest/minimal.html

from setuptools import find_packages, setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(
    name='fitgrid',
    version='0.1.2',
    description='Mass multiple regression manager',
    long_description=readme(),
    long_description_content_type='text/markdown',
    author='Thomas P. Urbach, Andrey Portnoy',
    author_email='turbach@ucsd.edu, aportnoy@ucsd.edu',
    url='https://github.com/kutaslab/fitgrid',
    license='BSD-3-Clause',
    packages=find_packages(exclude=['tests']),
    install_requires=['patsy', 'statsmodels', 'tqdm', 'tables'],
)



import numpy as np
import pandas as pd
from functools import lru_cache

from .errors import FitGridError


def add_epoch_index(temp, epoch_index):
    """We assume that temp is in long form, the columns are channels, and the
    first index level is TIME."""

    from . import TIME

    # first index level must be TIME
    assert temp.index.names[0] == TIME

    # temp should be long form, columns have single level (channels hopefully)
    assert not isinstance(temp.columns, pd.core.index.MultiIndex)

    # we can only handle 2- or 3-dimensional
    assert temp.index.nlevels in (2, 3)

    for i in range(1, temp.index.nlevels):
        level = temp.index.levels[i]
        # if a level looks like it was automatically created by Pandas,
        # we replace it with the epoch_index
        if (
            isinstance(level, pd.RangeIndex)
            and len(level) == len(epoch_index)
            and level._start == 0
            and level._step == 1
            and level._stop == len(epoch_index)
        ):
            temp.index.set_levels(epoch_index, level=i, inplace=True)
            temp.index.rename(epoch_index.name, level=i, inplace=True)

    return temp


def expand_series_or_df(temp):
    """Expand a DataFrame that has Series or DataFrames for values."""

    columns = (
        pd.concat(temp[channel].tolist(), keys=temp.index) for channel in temp
    )
    # concatenate columns, channel names are top level columns indices
    result = pd.concat(columns, axis=1, keys=temp.columns)

    # stack to achieve long form if columns have multiple levels
    if isinstance(result.columns, pd.core.indexes.multi.MultiIndex):
        return result.stack()
    return result


def _expand(temp, epoch_index):
    """Expand the values in the grid if possible, return frame or grid."""

    tester = temp.iloc[0, 0]

    # no processing needed
    if np.isscalar(tester):
        return temp

    # familiar types, expand them
    if isinstance(tester, pd.Series) or isinstance(tester, pd.DataFrame):
        # want single index level
        # can get more if original DataFrame had a multiindex
        # in Epochs we ensure that only EPOCH_ID is in the index for groupby
        if tester.index.nlevels > 1:
            raise NotImplementedError(
                f'index should have one level, have {tester.index.nlevels} '
                f'instead: {tester.index.names}'
            )
        return expand_series_or_df(temp)

    # array-like, try converting to array and then Series/DataFrame
    if isinstance(tester, tuple) or isinstance(tester, list):
        array_form = np.array(tester)
        if array_form.ndim == 1:
            temp = temp.applymap(lambda x: pd.Series(np.array(x)))
        elif array_form.ndim == 2:
            temp = temp.applymap(lambda x: pd.DataFrame(np.array(x)))
        else:
            raise NotImplementedError(
                'Cannot use elements with dim > 2,'
                f'element has ndim = {array_form.ndim}.'
            )
        temp_expanded = expand_series_or_df(temp)
        temp_epoch_index = add_epoch_index(temp_expanded, epoch_index)
        return temp_epoch_index

    # array, try converting to Series/DataFrame
    if isinstance(tester, np.ndarray):
        if tester.ndim == 1:
            temp = temp.applymap(lambda x: pd.Series(x))
        elif tester.ndim == 2:
            temp = temp.applymap(lambda x: pd.DataFrame(x))
        else:
            raise NotImplementedError(
                'Cannot use elements with dim > 2,'
                f'element has ndim = {tester.ndim}.'
            )
        temp_expanded = expand_series_or_df(temp)
        temp_epoch_index = add_epoch_index(temp_expanded, epoch_index)
        return temp_epoch_index

    # catchall for all types we don't handle explicitly
    # statsmodels objects, dicts, methods all go here
    return FitGrid(temp, epoch_index)


class FitGrid:
    """Hold rERP fit objects.

    FitGrid should not be created directly, the right way to build it is to
    start with an Epochs object and call a method like `.lm`.

    Parameters
    ----------
    _grid : Pandas DataFrame
        Pandas DataFrame of fit objects
    epochs_index : pandas Index
        index containing epoch ids

    Returns
    -------

    grid: FitGrid
        FitGrid object
    """

    def __init__(self, _grid, epoch_index):

        self._grid = _grid
        self._epoch_index = epoch_index
        self.tester = _grid.iloc[0, 0]
        self.channels = list(_grid.columns)

    def __getitem__(self, slicer):
        """Slice grid on time and channels, return new grid with that shape.

        The intended way to slice a FitGrid is to always slice on both time and
        channels, allowing wildcard colons:

            `grid[:, ['channel1', 'channel2']]`

        or

            `grid[:, :]`

        or

            `grid[25:-25, 'channel2']
        """

        if (
            not isinstance(slicer, tuple)
            or not hasattr(slicer, '__len__')
            or len(slicer) != 2
        ):
            raise FitGridError('Must slice on time and channels.')

        # now we can unpack
        time, channels = slicer

        def check_slicer_component(component):
            if isinstance(component, slice) or isinstance(component, list):
                return component
            else:
                # wrap in list to always get a DataFrame in return on slicing
                # otherwise we might get a scalar or a pandas Series,
                # which can't be used to create a FitGrid object
                return [component]

        time = check_slicer_component(time)
        channels = check_slicer_component(channels)
        subgrid = self._grid.loc[time, channels].copy()
        return self.__class__(subgrid, self._epoch_index)

    @lru_cache()
    def __getattr__(self, name):
        """Broadcast attribute extraction in the grid.

        This is the method that gets called when an attribute is requested that
        FitGrid does not have. We then go and check if the tester has it, and
        broadcast if it does.
        """

        if not hasattr(self.tester, name):
            raise FitGridError(f'No such attribute: {name}.')

        temp = self._grid.applymap(lambda x: getattr(x, name))
        return _expand(temp, self._epoch_index)

    def __call__(self, *args, **kwargs):
        """Broadcast method calling in the grid.

        This is the method that gets called when the grid is called as if it
        were a method.
        """
        if not callable(self.tester):
            raise FitGridError(
                f'This grid is not callable, '
                f'current type is {type(self.tester)}'
            )

        # if we are not callable, we'll get an appropriate exception
        temp = self._grid.applymap(lambda x: x(*args, **kwargs))
        return _expand(temp, self._epoch_index)

    def __dir__(self):

        # this exposes the fit object method but hides dunder methods, since
        # these likely overlap with the grid's own dunder methods
        return [item for item in dir(self.tester) if not item.startswith('__')]

    def __repr__(self):

        samples, chans = self._grid.shape
        return f'{samples} by {chans} FitGrid of type {type(self.tester)}.'

    def plot_betas(self):
        """Plot betas of the model, one plot per channel, overplotting betas.
        """

        if not hasattr(self.tester, 'params'):
            raise FitGridError('This FitGrid does not contain fit results.')

        import matplotlib.pyplot as plt

        with plt.rc_context({'font.size': 14}):
            params = self.params.unstack()
            channels, betas = params.columns.levels
            figsize = (16, 4 * len(channels))

            fig, axes = plt.subplots(
                nrows=len(channels), figsize=figsize, sharey=True
            )

            for ax, chan in zip(axes, channels):
                for beta in params[chan]:
                    ax.plot(params[chan][beta], label=beta)
                ax.set(ylabel=chan, xlabel=params.index.name)
                ax.legend(
                    loc='upper center',
                    ncol=len(params[chan]),
                    bbox_to_anchor=(0.5, 1.2),
                    fancybox=True,
                    shadow=False,
                )

            fig.tight_layout()

    def plot_adj_rsquared(self):
        """Plot adjusted :math:`R^2` as a heatmap with marginal bar and line.
        """

        import matplotlib.pyplot as plt

        if not hasattr(self.tester, 'rsquared_adj'):
            raise FitGridError('This FitGrid does not contain fit results.')

        rsq_adj = self.rsquared_adj

        with plt.rc_context({'font.size': 14}):
            plt.figure(figsize=(16, 12))
            gs = plt.GridSpec(2, 2, width_ratios=[13, 3], height_ratios=[7, 2])

            bar = plt.subplot(gs[1])
            bar.barh(self._grid.columns, rsq_adj.mean(axis=0))

            heatmap = plt.subplot(gs[0], sharey=bar)
            heatmap.imshow(rsq_adj.T, aspect='auto')
            heatmap.get_xaxis().set_visible(False)

            line = plt.subplot(gs[2])
            line.plot(rsq_adj.mean(axis=1))

            plt.tight_layout()
            plt.margins(x=0)

    def influential_epochs(self, top=None):
        """Return dataframe with top influential epochs ranked by Cook's-D.

        Parameters
        ----------
        top : int, optional, default None
            how many top epochs to return, all epochs by default

        Returns
        -------
        top_epochs : pandas DataFrame
            dataframe with EPOCH_ID as index and aggregated Cook's-D as values

        Notes
        -----
        Cook's distance is aggregated by simple averaging across time and
        channels.
        """

        if not hasattr(self.tester, 'get_influence'):
            raise FitGridError('This FitGrid does not contain influence data.')

        return (
            self.get_influence()
            .cooks_distance.drop(axis=0, index=1, level=1)
            .reset_index(level=1, drop=True)
            .mean(axis=0, level=1)
            .mean(axis=1)
            .sort_values(ascending=False)
            .to_frame(name='average_Cooks_D')
            .iloc[:top]
        )



import pandas as pd
from .epochs import Epochs
from .errors import FitGridError


def epochs_from_hdf(hdf_filename, channels='default'):
    """Construct Epochs object from an HDF5 file containing an epochs table.

    The HDF5 file should contain columns with names defined by EPOCH_ID and
    TIME either as index columns or as regular columns. This is added as a
    convenience, in general, input epochs tables should contain these columns
    in the index.

    Parameters
    ----------
    hdf_filename : str
        HDF5 file name
    channels : list of str, optional, defaults to CHANNELS
        list of string channel names

    Returns
    -------
    epochs : Epochs
        an Epochs object with the data
    """

    from . import EPOCH_ID, TIME

    df = pd.read_hdf(hdf_filename)

    # time and epoch id already present in index
    if EPOCH_ID in df.index.names and TIME in df.index.names:
        return Epochs(df, channels)

    # time and epoch id present in columns, set index
    if EPOCH_ID in df.columns and TIME in df.columns:
        df.set_index([EPOCH_ID, TIME], inplace=True)
        return Epochs(df, channels)

    raise FitGridError(
        f'Dataset has to contain {EPOCH_ID} and {TIME} as columns or indices.'
    )


def epochs_from_dataframe(dataframe, channels='default'):
    """Construct Epochs object from a Pandas DataFrame epochs table.

    The DataFrame should contain columns with names defined by EPOCH_ID and
    TIME as index columns.

    Parameters
    ----------
    dataframe : pandas DataFrame
        a pandas DataFrame object
    channels : list of str, optional, defaults to CHANNELS
        list of string channel names

    Returns
    -------
    epochs : Epochs
        an Epochs object with the data
    """
    return Epochs(dataframe, channels)



from .fake_data import generate
from .io import epochs_from_hdf, epochs_from_dataframe

#: default epoch identifier
EPOCH_ID = 'Epoch_idx'

#: default time identifier
TIME = 'Time'

#: default channel list
CHANNELS = [
    'lle',
    'lhz',
    'MiPf',
    'LLPf',
    'RLPf',
    'LMPf',
    'RMPf',
    'LDFr',
    'RDFr',
    'LLFr',
    'RLFr',
    'LMFr',
    'RMFr',
    'LMCe',
    'RMCe',
    'MiCe',
    'MiPa',
    'LDCe',
    'RDCe',
    'LDPa',
    'RDPa',
    'LMOc',
    'RMOc',
    'LLTe',
    'RLTe',
    'LLOc',
    'RLOc',
    'MiOc',
    'A2',
    'HEOG',
    'rle',
    'rhz',
]
